@@ -1,6 +1,6 @@
 
 /*
- * $Id: HttpHeader.c,v 1.74.2.25 2005/02/20 10:32:41 hno Exp $
+ * $Id: HttpHeader.c,v 1.99.2.1 2008/09/25 02:33:37 hno Exp $
  *
  * DEBUG: section 55    HTTP Header
  * AUTHOR: Alex Rousskov
@@ -80,14 +80,14 @@ static const HttpHeaderFieldAttrs HeadersAttrs[] =
     {"Content-Disposition", HDR_CONTENT_DISPOSITION, ftStr},
     {"Content-Encoding", HDR_CONTENT_ENCODING, ftStr},
     {"Content-Language", HDR_CONTENT_LANGUAGE, ftStr},
-    {"Content-Length", HDR_CONTENT_LENGTH, ftInt},
+    {"Content-Length", HDR_CONTENT_LENGTH, ftSize},
     {"Content-Location", HDR_CONTENT_LOCATION, ftStr},
     {"Content-MD5", HDR_CONTENT_MD5, ftStr},	/* for now */
     {"Content-Range", HDR_CONTENT_RANGE, ftPContRange},
     {"Content-Type", HDR_CONTENT_TYPE, ftStr},
     {"Cookie", HDR_COOKIE, ftStr},
     {"Date", HDR_DATE, ftDate_1123},
-    {"ETag", HDR_ETAG, ftETag},
+    {"ETag", HDR_ETAG, ftStr},
     {"Expires", HDR_EXPIRES, ftDate_1123},
     {"From", HDR_FROM, ftStr},
     {"Host", HDR_HOST, ftStr},
@@ -105,6 +105,7 @@ static const HttpHeaderFieldAttrs HeadersAttrs[] =
     {"Proxy-Authentication-Info", HDR_PROXY_AUTHENTICATION_INFO, ftStr},
     {"Proxy-Authorization", HDR_PROXY_AUTHORIZATION, ftStr},
     {"Proxy-Connection", HDR_PROXY_CONNECTION, ftStr},
+    {"Proxy-support", HDR_PROXY_SUPPORT, ftStr},
     {"Public", HDR_PUBLIC, ftStr},
     {"Range", HDR_RANGE, ftPRange},
     {"Referer", HDR_REFERER, ftStr},
@@ -112,12 +113,14 @@ static const HttpHeaderFieldAttrs HeadersAttrs[] =
     {"Retry-After", HDR_RETRY_AFTER, ftStr},	/* for now (ftDate_1123 or ftInt!) */
     {"Server", HDR_SERVER, ftStr},
     {"Set-Cookie", HDR_SET_COOKIE, ftStr},
-    {"Title", HDR_TITLE, ftStr},
     {"Transfer-Encoding", HDR_TRANSFER_ENCODING, ftStr},
+    {"Te", HDR_TE, ftStr},
+    {"Trailer", HDR_TRAILER, ftStr},
     {"Upgrade", HDR_UPGRADE, ftStr},	/* for now */
     {"User-Agent", HDR_USER_AGENT, ftStr},
     {"Vary", HDR_VARY, ftStr},	/* for now */
     {"Via", HDR_VIA, ftStr},	/* for now */
+    {"Expect", HDR_EXPECT, ftStr},
     {"Warning", HDR_WARNING, ftStr},	/* for now */
     {"WWW-Authenticate", HDR_WWW_AUTHENTICATE, ftStr},
     {"Authentication-Info", HDR_AUTHENTICATION_INFO, ftStr},
@@ -126,10 +129,15 @@ static const HttpHeaderFieldAttrs HeadersAttrs[] =
     {"X-Forwarded-For", HDR_X_FORWARDED_FOR, ftStr},
     {"X-Request-URI", HDR_X_REQUEST_URI, ftStr},
     {"X-Squid-Error", HDR_X_SQUID_ERROR, ftStr},
+    {"X-HTTP09-First-Line", HDR_X_HTTP09_FIRST_LINE, ftStr},
     {"Negotiate", HDR_NEGOTIATE, ftStr},
 #if X_ACCELERATOR_VARY
     {"X-Accelerator-Vary", HDR_X_ACCELERATOR_VARY, ftStr},
 #endif
+    {"X-Error-URL", HDR_X_ERROR_URL, ftStr},
+    {"X-Error-Status", HDR_X_ERROR_STATUS, ftInt},
+    {"Front-End-Https", HDR_FRONT_END_HTTPS, ftStr},
+    {"Keep-Alive", HDR_KEEP_ALIVE, ftStr},
     {"Other:", HDR_OTHER, ftStr}	/* ':' will not allow matches */
 };
 static HttpHeaderFieldInfo *Headers = NULL;
@@ -151,6 +159,9 @@ static http_hdr_type ListHeadersArr[] =
     HDR_IF_MATCH, HDR_IF_NONE_MATCH,
     HDR_LINK, HDR_PRAGMA,
     HDR_PROXY_CONNECTION,
+    HDR_PROXY_SUPPORT,
+    HDR_TE,
+    HDR_TRAILER,
     HDR_TRANSFER_ENCODING,
     HDR_UPGRADE,
     HDR_VARY,
@@ -159,7 +170,7 @@ static http_hdr_type ListHeadersArr[] =
     HDR_WWW_AUTHENTICATE,
     HDR_AUTHENTICATION_INFO,
     HDR_PROXY_AUTHENTICATION_INFO,
-    /* HDR_EXPECT, HDR_TE, HDR_TRAILER */
+    HDR_EXPECT,
 #if X_ACCELERATOR_VARY
     HDR_X_ACCELERATOR_VARY,
 #endif
@@ -171,8 +182,8 @@ static http_hdr_type GeneralHeadersArr[] =
 {
     HDR_CACHE_CONTROL, HDR_CONNECTION, HDR_DATE, HDR_PRAGMA,
     HDR_TRANSFER_ENCODING,
+    HDR_TRAILER,
     HDR_UPGRADE,
-    /* HDR_TRAILER, */
     HDR_VIA
 };
 
@@ -210,7 +221,7 @@ static http_hdr_type RequestHeadersArr[] =
     HDR_IF_MATCH, HDR_IF_MODIFIED_SINCE, HDR_IF_NONE_MATCH,
     HDR_IF_RANGE, HDR_MAX_FORWARDS, HDR_PROXY_CONNECTION,
     HDR_PROXY_AUTHORIZATION, HDR_RANGE, HDR_REFERER, HDR_REQUEST_RANGE,
-    HDR_USER_AGENT, HDR_X_FORWARDED_FOR
+    HDR_USER_AGENT, HDR_X_FORWARDED_FOR, HDR_TE, HDR_EXPECT
 };
 
 /* header accounting */
@@ -234,6 +245,7 @@ static int HeaderEntryParsedCount = 0;
 #define assert_eid(id) assert((id) < HDR_ENUM_END)
 
 static HttpHeaderEntry *httpHeaderEntryCreate(http_hdr_type id, const char *name, const char *value);
+static HttpHeaderEntry *httpHeaderEntryCreate2(http_hdr_type id, String name, String value);
 static void httpHeaderEntryDestroy(HttpHeaderEntry * e);
 static HttpHeaderEntry *httpHeaderEntryParseCreate(const char *field_start, const char *field_end);
 static void httpHeaderNoteParsedEntry(http_hdr_type id, String value, int error);
@@ -257,15 +269,15 @@ httpHeaderInitModule(void)
 	Headers = httpHeaderBuildFieldsInfo(HeadersAttrs, HDR_ENUM_END);
     /* create masks */
     httpHeaderMaskInit(&ListHeadersMask, 0);
-    httpHeaderCalcMask(&ListHeadersMask, (const int *) ListHeadersArr, countof(ListHeadersArr));
+    httpHeaderCalcMask(&ListHeadersMask, ListHeadersArr, countof(ListHeadersArr));
     httpHeaderMaskInit(&ReplyHeadersMask, 0);
-    httpHeaderCalcMask(&ReplyHeadersMask, (const int *) ReplyHeadersArr, countof(ReplyHeadersArr));
-    httpHeaderCalcMask(&ReplyHeadersMask, (const int *) GeneralHeadersArr, countof(GeneralHeadersArr));
-    httpHeaderCalcMask(&ReplyHeadersMask, (const int *) EntityHeadersArr, countof(EntityHeadersArr));
+    httpHeaderCalcMask(&ReplyHeadersMask, ReplyHeadersArr, countof(ReplyHeadersArr));
+    httpHeaderCalcMask(&ReplyHeadersMask, GeneralHeadersArr, countof(GeneralHeadersArr));
+    httpHeaderCalcMask(&ReplyHeadersMask, EntityHeadersArr, countof(EntityHeadersArr));
     httpHeaderMaskInit(&RequestHeadersMask, 0);
-    httpHeaderCalcMask(&RequestHeadersMask, (const int *) RequestHeadersArr, countof(RequestHeadersArr));
-    httpHeaderCalcMask(&RequestHeadersMask, (const int *) GeneralHeadersArr, countof(GeneralHeadersArr));
-    httpHeaderCalcMask(&RequestHeadersMask, (const int *) EntityHeadersArr, countof(EntityHeadersArr));
+    httpHeaderCalcMask(&RequestHeadersMask, RequestHeadersArr, countof(RequestHeadersArr));
+    httpHeaderCalcMask(&RequestHeadersMask, GeneralHeadersArr, countof(GeneralHeadersArr));
+    httpHeaderCalcMask(&RequestHeadersMask, EntityHeadersArr, countof(EntityHeadersArr));
     /* init header stats */
     assert(HttpHeaderStatCount == hoReply + 1);
     for (i = 0; i < HttpHeaderStatCount; i++)
@@ -366,16 +378,39 @@ httpHeaderAppend(HttpHeader * dest, const HttpHeader * src)
     debug(55, 7) ("appending hdr: %p += %p\n", dest, src);
 
     while ((e = httpHeaderGetEntry(src, &pos))) {
-	httpHeaderAddEntry(dest, httpHeaderEntryClone(e));
+	httpHeaderAddClone(dest, e);
     }
 }
 
+static void
+httpHeaderRepack(HttpHeader * hdr)
+{
+    HttpHeaderPos dp = HttpHeaderInitPos;
+    HttpHeaderPos pos = HttpHeaderInitPos;
+
+    /* XXX breaks layering for now! ie, getting grubby fingers in without httpHeaderEntryGet() */
+    dp = 0;
+    pos = 0;
+    while (dp < hdr->entries.count) {
+	for (; dp < hdr->entries.count && hdr->entries.items[dp] == NULL; dp++);
+	if (dp >= hdr->entries.count)
+	    break;
+	hdr->entries.items[pos] = hdr->entries.items[dp];
+	if (dp != pos)
+	    hdr->entries.items[dp] = NULL;
+	pos++;
+	dp++;
+    }
+    arrayShrink(&hdr->entries, pos);
+}
+
 /* use fresh entries to replace old ones */
 void
 httpHeaderUpdate(HttpHeader * old, const HttpHeader * fresh, const HttpHeaderMask * denied_mask)
 {
     const HttpHeaderEntry *e;
     HttpHeaderPos pos = HttpHeaderInitPos;
+
     assert(old && fresh);
     assert(old != fresh);
     debug(55, 7) ("updating hdr: %p <- %p\n", old, fresh);
@@ -384,9 +419,21 @@ httpHeaderUpdate(HttpHeader * old, const HttpHeader * fresh, const HttpHeaderMas
 	/* deny bad guys (ok to check for HDR_OTHER) here */
 	if (denied_mask && CBIT_TEST(*denied_mask, e->id))
 	    continue;
-	httpHeaderDelByName(old, strBuf(e->name));
-	httpHeaderAddEntry(old, httpHeaderEntryClone(e));
+	if (e->id != HDR_OTHER)
+	    httpHeaderDelById(old, e->id);
+	else
+	    httpHeaderDelByName(old, strBuf(e->name));
     }
+    pos = HttpHeaderInitPos;
+    while ((e = httpHeaderGetEntry(fresh, &pos))) {
+	/* deny bad guys (ok to check for HDR_OTHER) here */
+	if (denied_mask && CBIT_TEST(*denied_mask, e->id))
+	    continue;
+	httpHeaderAddClone(old, e);
+    }
+
+    /* And now, repack the array to "fill in the holes" */
+    httpHeaderRepack(old);
 }
 
 /* just handy in parsing: resets and returns false */
@@ -405,7 +452,7 @@ int
 httpHeaderParse(HttpHeader * hdr, const char *header_start, const char *header_end)
 {
     const char *field_ptr = header_start;
-    HttpHeaderEntry *e, *e2;
+    HttpHeaderEntry *e;
 
     assert(hdr);
     assert(header_start && header_end);
@@ -473,14 +520,43 @@ httpHeaderParse(HttpHeader * hdr, const char *header_start, const char *header_e
 	    else
 		return httpHeaderReset(hdr);
 	}
-	if (e->id == HDR_CONTENT_LENGTH && (e2 = httpHeaderFindEntry(hdr, e->id)) != NULL) {
-	    if (!Config.onoff.relaxed_header_parser || strCmp(e->value, strBuf(e2->value)) != 0) {
-		debug(55, 1) ("WARNING: found two conflicting content-length headers in {%s}\n", getStringPrefix(header_start, header_end));
+	if (e->id == HDR_CONTENT_LENGTH) {
+	    squid_off_t l1;
+	    HttpHeaderEntry *e2;
+	    if (!httpHeaderParseSize(strBuf(e->value), &l1)) {
+		debug(55, 1) ("WARNING: Unparseable content-length '%s'\n", strBuf(e->value));
 		httpHeaderEntryDestroy(e);
 		return httpHeaderReset(hdr);
-	    } else {
+	    }
+	    e2 = httpHeaderFindEntry(hdr, e->id);
+	    if (e2 && strCmp(e->value, strBuf(e2->value)) != 0) {
+		squid_off_t l2;
+		debug(55, Config.onoff.relaxed_header_parser <= 0 ? 1 : 2) ("WARNING: found two conflicting content-length headers in {%s}\n", getStringPrefix(header_start, header_end));
+		if (!Config.onoff.relaxed_header_parser) {
+		    httpHeaderEntryDestroy(e);
+		    return httpHeaderReset(hdr);
+		}
+		if (!httpHeaderParseSize(strBuf(e2->value), &l2)) {
+		    debug(55, 1) ("WARNING: Unparseable content-length '%s'\n", strBuf(e->value));
+		    httpHeaderEntryDestroy(e);
+		    return httpHeaderReset(hdr);
+		}
+		if (l1 > l2) {
+		    httpHeaderDelById(hdr, e2->id);
+		} else {
+		    httpHeaderEntryDestroy(e);
+		    continue;
+		}
+	    } else if (e2) {
 		debug(55, Config.onoff.relaxed_header_parser <= 0 ? 1 : 2)
 		    ("NOTICE: found double content-length header\n");
+		if (Config.onoff.relaxed_header_parser) {
+		    httpHeaderEntryDestroy(e);
+		    continue;
+		} else {
+		    httpHeaderEntryDestroy(e);
+		    return httpHeaderReset(hdr);
+		}
 	    }
 	}
 	if (e->id == HDR_OTHER && stringHasWhitespace(strBuf(e->name))) {
@@ -622,6 +698,8 @@ httpHeaderDelById(HttpHeader * hdr, http_hdr_type id)
 /*
  * deletes an entry at pos and leaves a gap; leaving a gap makes it
  * possible to iterate(search) and delete fields at the same time
+ * WARNING: Doesn't update the header mask. Call httpHeaderRefreshMask
+ * when done with the delete operations.
  */
 void
 httpHeaderDelAt(HttpHeader * hdr, HttpHeaderPos pos)
@@ -636,6 +714,22 @@ httpHeaderDelAt(HttpHeader * hdr, HttpHeaderPos pos)
     httpHeaderEntryDestroy(e);
 }
 
+/*
+ * Refreshes the header mask. Useful after httpHeaderDelAt constructs
+ */
+void
+httpHeaderRefreshMask(HttpHeader * hdr)
+{
+    HttpHeaderPos pos = HttpHeaderInitPos;
+    HttpHeaderEntry *e;
+    httpHeaderMaskInit(&hdr->mask, 0);
+    debug(55, 7) ("refreshing the mask in hdr %p\n", hdr);
+    while ((e = httpHeaderGetEntry(hdr, &pos))) {
+	CBIT_SET(hdr->mask, e->id);
+    }
+}
+
+
 
 /* appends an entry; 
  * does not call httpHeaderEntryClone() so one should not reuse "*e"
@@ -657,6 +751,26 @@ httpHeaderAddEntry(HttpHeader * hdr, HttpHeaderEntry * e)
     hdr->len += strLen(e->name) + 2 + strLen(e->value) + 2;
 }
 
+/* inserts an entry at the given position; 
+ * does not call httpHeaderEntryClone() so one should not reuse "*e"
+ */
+void
+httpHeaderInsertEntry(HttpHeader * hdr, HttpHeaderEntry * e, int pos)
+{
+    assert(hdr && e);
+    assert_eid(e->id);
+
+    debug(55, 7) ("%p adding entry: %d at %d\n",
+	hdr, e->id, hdr->entries.count);
+    if (CBIT_TEST(hdr->mask, e->id))
+	Headers[e->id].stat.repCount++;
+    else
+	CBIT_SET(hdr->mask, e->id);
+    arrayInsert(&hdr->entries, e, pos);
+    /* increment header length, allow for ": " and crlf */
+    hdr->len += strLen(e->name) + 2 + strLen(e->value) + 2;
+}
+
 /* return a list of entries with the same id separated by ',' and ws */
 String
 httpHeaderGetList(const HttpHeader * hdr, http_hdr_type id)
@@ -695,7 +809,7 @@ httpHeaderGetStrOrList(const HttpHeader * hdr, http_hdr_type id)
 	return httpHeaderGetList(hdr, id);
     if ((e = httpHeaderFindEntry(hdr, id))) {
 	String s;
-	stringInit(&s, strBuf(e->value));
+	stringLimitInit(&s, strBuf(e->value), strLen(e->value));
 	return s;
     }
     return StringNull;
@@ -755,6 +869,7 @@ httpHeaderGetByNameListMember(const HttpHeader * hdr, const char *name, const ch
 	    break;
 	}
     }
+    stringClean(&header);
     return result;
 }
 
@@ -806,6 +921,17 @@ httpHeaderPutInt(HttpHeader * hdr, http_hdr_type id, int number)
     httpHeaderAddEntry(hdr, httpHeaderEntryCreate(id, NULL, xitoa(number)));
 }
 
+void
+httpHeaderPutSize(HttpHeader * hdr, http_hdr_type id, squid_off_t number)
+{
+    char size[64];
+    assert_eid(id);
+    assert(Headers[id].type == ftSize);		/* must be of an appropriate type */
+    assert(number >= 0);
+    snprintf(size, sizeof(size), "%" PRINTF_OFF_T, number);
+    httpHeaderAddEntry(hdr, httpHeaderEntryCreate(id, NULL, size));
+}
+
 void
 httpHeaderPutTime(HttpHeader * hdr, http_hdr_type id, time_t htime)
 {
@@ -815,6 +941,15 @@ httpHeaderPutTime(HttpHeader * hdr, http_hdr_type id, time_t htime)
     httpHeaderAddEntry(hdr, httpHeaderEntryCreate(id, NULL, mkrfc1123(htime)));
 }
 
+void
+httpHeaderInsertTime(HttpHeader * hdr, int pos, http_hdr_type id, time_t htime)
+{
+    assert_eid(id);
+    assert(Headers[id].type == ftDate_1123);	/* must be of an appropriate type */
+    assert(htime >= 0);
+    httpHeaderInsertEntry(hdr, httpHeaderEntryCreate(id, NULL, mkrfc1123(htime)), pos);
+}
+
 void
 httpHeaderPutStr(HttpHeader * hdr, http_hdr_type id, const char *str)
 {
@@ -912,6 +1047,21 @@ httpHeaderGetInt(const HttpHeader * hdr, http_hdr_type id)
     return value;
 }
 
+squid_off_t
+httpHeaderGetSize(const HttpHeader * hdr, http_hdr_type id)
+{
+    HttpHeaderEntry *e;
+    squid_off_t value = -1;
+    int ok;
+    assert_eid(id);
+    assert(Headers[id].type == ftSize);		/* must be of an appropriate type */
+    if ((e = httpHeaderFindEntry(hdr, id))) {
+	ok = httpHeaderParseSize(strBuf(e->value), &value);
+	httpHeaderNoteParsedEntry(e->id, e->value, !ok);
+    }
+    return value;
+}
+
 time_t
 httpHeaderGetTime(const HttpHeader * hdr, http_hdr_type id)
 {
@@ -920,7 +1070,7 @@ httpHeaderGetTime(const HttpHeader * hdr, http_hdr_type id)
     assert_eid(id);
     assert(Headers[id].type == ftDate_1123);	/* must be of an appropriate type */
     if ((e = httpHeaderFindEntry(hdr, id))) {
-	value = parse_rfc1123(strBuf(e->value));
+	value = parse_rfc1123(strBuf(e->value), strLen(e->value));
 	httpHeaderNoteParsedEntry(e->id, e->value, value < 0);
     }
     return value;
@@ -1022,18 +1172,6 @@ httpHeaderGetAuth(const HttpHeader * hdr, http_hdr_type id, const char *auth_sch
     return base64_decode(field);
 }
 
-ETag
-httpHeaderGetETag(const HttpHeader * hdr, http_hdr_type id)
-{
-    ETag etag =
-    {NULL, -1};
-    HttpHeaderEntry *e;
-    assert(Headers[id].type == ftETag);		/* must be of an appropriate type */
-    if ((e = httpHeaderFindEntry(hdr, id)))
-	etagParseInit(&etag, strBuf(e->value));
-    return etag;
-}
-
 TimeOrTag
 httpHeaderGetTimeOrTag(const HttpHeader * hdr, http_hdr_type id)
 {
@@ -1044,17 +1182,20 @@ httpHeaderGetTimeOrTag(const HttpHeader * hdr, http_hdr_type id)
     if ((e = httpHeaderFindEntry(hdr, id))) {
 	const char *str = strBuf(e->value);
 	/* try as an ETag */
-	if (etagParseInit(&tot.tag, str)) {
-	    tot.valid = tot.tag.str != NULL;
+	if (*str == '"' || (str[0] == 'W' && str[1] == '/')) {
+	    tot.tag = str;
 	    tot.time = -1;
+	    tot.valid = 1;
 	} else {
 	    /* or maybe it is time? */
-	    tot.time = parse_rfc1123(str);
-	    tot.valid = tot.time >= 0;
-	    tot.tag.str = NULL;
+	    tot.time = parse_rfc1123(str, strLen(e->value));
+	    if (tot.time >= 0)
+		tot.valid = 1;
+	    tot.tag = NULL;
 	}
+    } else {
+	tot.time = -1;
     }
-    assert(tot.time < 0 || !tot.tag.str);	/* paranoid */
     return tot;
 }
 
@@ -1079,6 +1220,23 @@ httpHeaderEntryCreate(http_hdr_type id, const char *name, const char *value)
     return e;
 }
 
+static HttpHeaderEntry *
+httpHeaderEntryCreate2(http_hdr_type id, String name, String value)
+{
+    HttpHeaderEntry *e;
+    assert_eid(id);
+    e = memAllocate(MEM_HTTP_HDR_ENTRY);
+    e->id = id;
+    if (id != HDR_OTHER)
+	e->name = Headers[id].name;
+    else
+	stringLimitInit(&e->name, strBuf(name), strLen(name));
+    stringLimitInit(&e->value, strBuf(value), strLen(value));
+    Headers[id].stat.aliveCount++;
+    debug(55, 9) ("created entry %p: '%s: %s'\n", e, strBuf(e->name), strBuf(e->value));
+    return e;
+}
+
 static void
 httpHeaderEntryDestroy(HttpHeaderEntry * e)
 {
@@ -1112,8 +1270,8 @@ httpHeaderEntryParseCreate(const char *field_start, const char *field_end)
     /* do we have a valid field name within this field? */
     if (!name_len || name_end > field_end)
 	return NULL;
-    if (name_len > 65536) {
-	/* String has a 64K limit */
+    if (name_len > 65534) {
+	/* String must be LESS THAN 64K and it adds a terminating NULL */
 	debug(55, 1) ("WARNING: ignoring header name of %d bytes\n", name_len);
 	return NULL;
     }
@@ -1144,8 +1302,8 @@ httpHeaderEntryParseCreate(const char *field_start, const char *field_end)
 	value_start++;
     while (value_start < field_end && xisspace(field_end[-1]))
 	field_end--;
-    if (field_end - value_start > 65536) {
-	/* String has a 64K limit */
+    if (field_end - value_start > 65534) {
+	/* String must be LESS THAN 64K and it adds a terminating NULL */
 	debug(55, 1) ("WARNING: ignoring '%s' header of %d bytes\n",
 	    strBuf(e->name), (int) (field_end - value_start));
 	if (e->id == HDR_OTHER)
@@ -1164,7 +1322,13 @@ httpHeaderEntryParseCreate(const char *field_start, const char *field_end)
 HttpHeaderEntry *
 httpHeaderEntryClone(const HttpHeaderEntry * e)
 {
-    return httpHeaderEntryCreate(e->id, strBuf(e->name), strBuf(e->value));
+    return httpHeaderEntryCreate2(e->id, e->name, e->value);
+}
+
+void
+httpHeaderAddClone(HttpHeader * hdr, const HttpHeaderEntry * e)
+{
+    httpHeaderAddEntry(hdr, httpHeaderEntryClone(e));
 }
 
 void
